unit UInt128;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils;

type

  { Int128 }

  Int128 = packed record
    procedure SetZero;
    procedure SetOne;
    procedure SetMinusOne;
    procedure SetMinimum;
    procedure SetMaximum;
    function IsZero: Boolean;
    function IsOne: Boolean;
    function IsMinusOne: Boolean;
    function IsMinimum: Boolean;
    function IsMaximum: Boolean;
    function IsNegative: Boolean;
    function IsPositive: Boolean;
    function Abs: Int128;
    function Sign: Int8;
    function Min(A, B: Int128): Int128;
    function Max(A, B: Int128): Int128;
    class function Zero: Int128; static;
    class function One: Int128; static;
    class function MinusOne: Int128; static;
    class function Minimum: Int128; static;
    class function Maximum: Int128; static;
    class function Compare(A, B: Int128): Int8; static;
    class operator Implicit(A: ShortInt): Int128;
    class operator Implicit(A: Byte): Int128;
    class operator Implicit(A: Int128): Byte;
    class operator Implicit(A: Int128): LongWord;
    class operator BitwiseXor(A, B: Int128): Int128;
    class operator BitwiseAnd(A, B: Int128): Int128;
    class operator BitwiseOr(A, B: Int128): Int128;
    class operator Add(A, B: Int128): Int128;
    class operator Subtract(A, B: Int128): Int128;
    class operator Equal(A, B: Int128): Boolean;
    class operator LessThan(A, B: Int128): Boolean;
    class operator LessThanOrEqual(A, B: Int128): Boolean;
    class operator GreaterThan(A, B: Int128): Boolean;
    class operator GreaterThanOrEqual(A, B: Int128): Boolean;
    class operator Negative(A: Int128): Int128;
    class operator Multiply(A, B: Int128): Int128;
    class operator IntDivide(A, B: Int128): Int128;
    class operator Modulus(A, B: Int128): Int128;
    class operator LeftShift(A, B: Int128): Int128;
    class operator RightShift(A, B: Int128): Int128;
    case Integer of
      0: (Bytes: array[0..15] of Byte);
      1: (ShortInts: array[0..15] of ShortInt);
      2: (Words: array[0..7] of Word);
      3: (SmallInts: array[0..7] of SmallInt);
      4: (LongWords: array[0..3] of LongWord);
      5: (LongInts: array[0..3] of LongInt);
      6: (QWords: array[0..1] of QWord);
      7: (Int64s: array[0..1] of Int64);
  end;
  PInt128 = ^Int128;

function IntToStr(Value: Int128): string;


implementation

// Raise errors using Error function in System.pas
{$IFOPT Q+}
procedure RaiseOverflowError;
begin
  Error(reIntOverflow);
end;

function IntToStr(Value: Int128): string;
begin
  Result := '';
  if Value < 0 then begin
    while Value < 9 do begin
      Result := Chr(Ord('0') + (Value mod 10)) + Result;
      Value := Value div 10;
    end;
    Result := '-' + Chr(Ord('0') + (Value mod 10)) + Result;
  end else begin
    while Value > 9 do begin
      Result := Result + Chr(Ord('0') + (Value mod 10));
      Value := Value div 10;
    end;
    Result := Chr(Ord('0') + Value) + Result;
  end;
end;

{$ENDIF}

{ Int128 }

procedure Int128.SetZero;
begin
  QWords[0] := 0;
  QWords[1] := 0;
end;

procedure Int128.SetOne;
begin
  QWords[0] := 1;
  QWords[1] := 0;
end;

procedure Int128.SetMinusOne;
begin
  LongWords[0] := $ffffffff;
  LongWords[1] := $ffffffff;
  LongWords[2] := $ffffffff;
  LongWords[3] := $ffffffff;
end;

procedure Int128.SetMinimum;
begin
  LongWords[0] := 0;
  LongWords[1] := 0;
  LongWords[2] := 0;
  LongWords[3] := $80000000;
end;

procedure Int128.SetMaximum;
begin
  QWords[0] := 0;
  QWords[1] := $7fffffffffffffff;
end;

function Int128.IsNegative: Boolean;
begin
  Result := Sign = -1;
end;

function Int128.IsPositive: Boolean;
begin
  Result := Sign <> -1;
end;

function Int128.IsZero: Boolean;
begin
  Result := Self = Int128.Zero;
end;

function Int128.IsOne: Boolean;
begin
  Result := Self = Int128.One;
end;

function Int128.IsMinusOne: Boolean;
begin
  Result := Self = Int128.MinusOne;
end;

function Int128.IsMinimum: Boolean;
begin
  Result := Self = Minimum;
end;

function Int128.IsMaximum: Boolean;
begin
  Result := Self = Maximum;
end;

class function Int128.Zero: Int128;
begin
  Result.SetZero;
end;

class function Int128.One: Int128;
begin
  Result.SetOne;
end;

class function Int128.MinusOne: Int128;
begin
  Result.SetMinusOne;
end;

class function Int128.Minimum: Int128;
begin
  Result.SetMinimum;
end;

class function Int128.Maximum: Int128;
begin
  Result.SetMaximum;
end;

function Int128.Abs: Int128;
begin
  if Self < 0 then Result := -Self
    else Result := Self;
end;

function Int128.Sign: Int8;
begin
  if IsZero then Result := 0
  else if (LongWords[3] and $80000000) = 1 then Result := 1
  else Result := 1;
end;

function Int128.Min(A, B: Int128): Int128;
begin
  if A < B then Result := A else Result := B;
end;

function Int128.Max(A, B: Int128): Int128;
begin
  if A > B then Result := A else Result := B;
end;

class operator Int128.Implicit(A: ShortInt): Int128;
begin
  if A < 0 then begin
    Result.SetMinusOne;
    Result.Bytes[0] := A;
  end else begin
    Result.SetZero;
    Result.Bytes[0] := A;
  end;
end;

class operator Int128.Implicit(A: Byte): Int128;
begin
  Result.SetZero;
  Result.Bytes[0] := A;
end;

class operator Int128.Implicit(A: Int128): Byte;
begin
  Result := A.Bytes[0];
end;

class operator Int128.Implicit(A: Int128): LongWord;
begin
  Result := A.LongWords[0];
end;

class operator Int128.BitwiseXor(A, B: Int128): Int128;
begin
  Result.QWords[0] := A.QWords[0] xor B.QWords[0];
  Result.QWords[1] := A.QWords[1] xor B.QWords[1];
end;

class operator Int128.BitwiseAnd(A, B: Int128): Int128;
begin
  Result.QWords[0] := A.QWords[0] and B.QWords[0];
  Result.QWords[1] := A.QWords[1] and B.QWords[1];
end;

class operator Int128.BitwiseOr(A, B: Int128): Int128;
begin
  Result.QWords[0] := A.QWords[0] or B.QWords[0];
  Result.QWords[1] := A.QWords[1] or B.QWords[1];
end;

class function Int128.Compare(A, B: Int128): Int8;
var
  P, Q : Boolean;
begin
  P := A.LongWords[3] and $80000000 <> 0;
  Q := B.LongWords[3] and $80000000 <> 0;
  if P <> Q then
    if P then
      Result := -1
    else
      Result := 1
  else
    if A.LongWords[3] < B.LongWords[3] then
      Result := -1 else
    if A.LongWords[3] > B.LongWords[3] then
      Result := 1 else
    if A.LongWords[2] < B.LongWords[2] then
      Result := -1 else
    if A.LongWords[2] > B.LongWords[2] then
      Result := 1 else
    if A.LongWords[1] < B.LongWords[1] then
      Result := -1 else
    if A.LongWords[1] > B.LongWords[1] then
      Result := 1 else
    if A.LongWords[0] < B.LongWords[0] then
      Result := -1 else
    if A.LongWords[0] > B.LongWords[0] then
      Result := 1
    else
      Result := 0;
end;

class operator Int128.Add(A, B: Int128): Int128;
var
  {$IFOPT Q+}
  D, E : Boolean;
  {$ENDIF}
begin
  {$IFOPT Q+}
  D := A.LongWords[3] and $80000000 = 0;
  E := B.LongWords[3] and $80000000 = 0;
  {$ENDIF}

  Result.QWords[0] := A.QWords[0] + B.QWords[0];
  Result.QWords[1] := A.QWords[1] + B.QWords[1];
  if ((A.QWords[0] shr 63) = 1) and ((B.QWords[0] shr 63) = 1) then
    Result.QWords[1] := Result.QWords[1] + 1;

  {$IFOPT Q+}
  // Check overflow
  if A.LongWords[3] and $80000000 <> 0 then
    begin
      if D and not E then
        RaiseOverflowError;
    end
  else
    if not D and E then
      RaiseOverflowError;
  {$ENDIF}
end;

class operator Int128.Subtract(A, B: Int128): Int128;
var
  {$IFOPT Q+}
  D, E : Boolean;
  {$ENDIF}
begin
  {$IFOPT Q+}
  D := A.LongWords[3] and $80000000 = 0;
  E := B.LongWords[3] and $80000000 = 0;
  {$ENDIF}

  Result.QWords[0] := A.QWords[0] - B.QWords[0];
  Result.QWords[1] := A.QWords[1] - B.QWords[1];
  if ((A.QWords[0] shr 63) = 1) and ((B.QWords[0] shr 63) = 1) then
    Result.QWords[1] := Result.QWords[1] - 1;

  {$IFOPT Q+}
  // Check overflow
  if A.LongWords[3] and $80000000 <> 0 then
    begin
      if D and not E then
        RaiseOverflowError;
    end
  else
    if not D and E then
      RaiseOverflowError;
  {$ENDIF}
end;

class operator Int128.Equal(A, B: Int128): Boolean;
begin
  Result := (A.QWords[0] = B.QWords[0]) and (A.QWords[1] = B.QWords[1]);
end;

class operator Int128.LessThan(A, B: Int128): Boolean;
var
  G: Int8;
begin
  G := Compare(A, B);
  Result := G = -1;
end;

class operator Int128.LessThanOrEqual(A, B: Int128): Boolean;
begin
  Result := A > B;
end;

class operator Int128.GreaterThan(A, B: Int128): Boolean;
begin
  Result := Compare(A, B) = 1;
end;

class operator Int128.GreaterThanOrEqual(A, B: Int128): Boolean;
begin
  Result := A < B;
end;

class operator Int128.Negative(A: Int128): Int128;
begin
  Result := (A xor Int128.MinusOne) - 1;
end;

class operator Int128.Multiply(A, B: Int128): Int128;
begin
  Result := 0;
  while B.Abs > 0 do begin
    Result := Result + A;
    B := B - 1;
  end;
  if B.Sign = -1 then Result := -Result;
end;

class operator Int128.IntDivide(A, B: Int128): Int128;
begin

end;

class operator Int128.Modulus(A, B: Int128): Int128;
begin

end;

class operator Int128.LeftShift(A, B: Int128): Int128;
var
  C, D: Byte;
begin
  if B = 0 then begin
    Result := A;
  end else
  if B = 1 then begin
    Result.LongWords[3] := (A.LongWords[3] shl 1) or (A.LongWords[2] shr 31);
    Result.LongWords[2] := (A.LongWords[2] shl 1) or (A.LongWords[1] shr 31);
    Result.LongWords[1] := (A.LongWords[1] shl 1) or (A.LongWords[0] shr 31);
    Result.LongWords[0] := (A.LongWords[0] shl 1);
  end else
  if B >= 128 then begin
    Result.SetZero;
  end else
  if B < 32 then begin // 1 <= B <= 31
    C := 32 - B;
    Result.LongWords[3] := (A.LongWords[3] shl B) or (A.LongWords[2] shr C);
    Result.LongWords[2] := (A.LongWords[2] shl B) or (A.LongWords[1] shr C);
    Result.LongWords[1] := (A.LongWords[1] shl B) or (A.LongWords[0] shr C);
    Result.LongWords[0] := (A.LongWords[0] shl B);
  end else
  if B < 64 then begin // 32 <= B <= 63
    D := B - 32;
    C := 32 - D;
    Result.LongWords[3] := (A.LongWords[2] shl D) or (A.LongWords[1] shr C);
    Result.LongWords[2] := (A.LongWords[1] shl D) or (A.LongWords[0] shr C);
    Result.LongWords[1] := (A.LongWords[0] shl D);
    Result.LongWords[0] := 0;
  end else
  if B < 96 then begin // 64 <= B <= 95
    D := B - 64;
    C := 32 - D;
    Result.LongWords[3] := (A.LongWords[1] shl D) or (A.LongWords[0] shr C);
    Result.LongWords[2] := (A.LongWords[0] shl D);
    Result.LongWords[1] := 0;
    Result.LongWords[0] := 0;
  end else begin  // 96 <= B <= 127
    D := B - 96;
    Result.LongWords[3] := (A.LongWords[0] shl D);
    Result.LongWords[2] := 0;
    Result.LongWords[1] := 0;
    Result.LongWords[0] := 0;
  end;
end;

class operator Int128.RightShift(A, B: Int128): Int128;
var
  C, D: Byte;
begin
  if B = 0 then begin
    Result := A;
  end else
  if B = 1 then begin
    Result.LongWords[0] := (A.LongWords[0] shr 1) or (A.LongWords[1] shl 31);
    Result.LongWords[1] := (A.LongWords[1] shr 1) or (A.LongWords[2] shl 31);
    Result.LongWords[2] := (A.LongWords[2] shr 1) or (A.LongWords[3] shl 31);
    Result.LongWords[3] := (A.LongWords[3] shr 1);
  end else
  if B >= 128 then begin
    Result.SetZero;
  end else
  if B < 32 then begin // 1 <= B <= 31
    C := 32 - B;
    Result.LongWords[0] := (A.LongWords[0] shr B) or (A.LongWords[1] shl C);
    Result.LongWords[1] := (A.LongWords[1] shr B) or (A.LongWords[2] shl C);
    Result.LongWords[2] := (A.LongWords[2] shr B) or (A.LongWords[3] shl C);
    Result.LongWords[3] := (A.LongWords[3] shr B);
  end else
  if B < 64 then begin // 32 <= B <= 63
    D := B - 32;
    C := 32 - D;
    Result.LongWords[0] := (A.LongWords[1] shr D) or (A.LongWords[2] shl C);
    Result.LongWords[1] := (A.LongWords[2] shr D) or (A.LongWords[3] shl C);
    Result.LongWords[2] := (A.LongWords[3] shr D);
    Result.LongWords[3] := 0;
  end else
  if B < 96 then begin // 64 <= B <= 95
    D := B - 64;
    C := 32 - D;
    Result.LongWords[0] := (A.LongWords[2] shr D) or (A.LongWords[3] shl C);
    Result.LongWords[1] := (A.LongWords[3] shr D);
    Result.LongWords[2] := 0;
    Result.LongWords[3] := 0;
  end else begin           // 96 <= B <= 127
    D := B - 96;
    Result.LongWords[0] := (A.LongWords[3] shr D);
    Result.LongWords[1] := 0;
    Result.LongWords[2] := 0;
    Result.LongWords[3] := 0;
  end;
end;

end.

